import pytest
import json
from unittest.mock import patch, MagicMock
from kage.connectors.discord import DiscordConnector
from kage.connectors.telegram import TelegramConnector
from kage.config import DiscordConnectorConfig, TelegramConnectorConfig

@patch("kage.connectors.discord.urllib.request.urlopen")
@patch("kage.connectors.discord.generate_chat_reply")
def test_discord_connector_poll_and_reply(mock_chat, mock_urlopen, tmp_path):
    config = DiscordConnectorConfig(
        active=True, bot_token="token", channel_id="123"
    )
    connector = DiscordConnector("test_discord", config)
    connector.state_file = tmp_path / "discord_state.json"

    # Recent timestamp so the message passes max_age_seconds filter
    from datetime import datetime, timezone
    recent_ts = datetime.now(timezone.utc).isoformat()

    # Mock API: _get_bot_identity call
    mock_identity_response = MagicMock()
    mock_identity_response.__enter__.return_value = mock_identity_response
    mock_identity_response.read.return_value = json.dumps(
        {"id": "999", "username": "kage"}
    ).encode("utf-8")

    # Mock API: GET messages
    mock_get_response = MagicMock()
    mock_get_response.__enter__.return_value = mock_get_response
    mock_get_response.read.return_value = json.dumps([
        {"id": "1", "content": "hello kage", "author": {"bot": False}, "timestamp": recent_ts}
    ]).encode("utf-8")
    
    # Mock API: POST reply (returns message id "2")
    mock_post_response = MagicMock()
    mock_post_response.__enter__.return_value = mock_post_response
    mock_post_response.read.return_value = json.dumps({"id": "2"}).encode("utf-8")
    
    # urlopen calls: GET messages, _get_bot_identity, POST reply
    mock_urlopen.side_effect = [mock_get_response, mock_identity_response, mock_post_response]

    # Mock AI reply
    mock_chat.return_value = {"stdout": "hello human", "stderr": "", "returncode": 0}

    connector.poll_and_reply()

    # Check state file updated to bot's reply ID
    assert connector.state_file.exists()
    state = json.loads(connector.state_file.read_text())
    assert state["last_message_id"] == "2"

    # AI Chat triggered correctly
    mock_chat.assert_called_once()
    actual_prompt = mock_chat.call_args[0][0]
    assert "[Recent Chat History]" in actual_prompt
    assert "[Current Instruction]\nhello kage" in actual_prompt

@patch("kage.connectors.discord.urllib.request.urlopen")
def test_discord_connector_ignores_bot_messages(mock_urlopen, tmp_path):
    config = DiscordConnectorConfig(
        active=True, bot_token="token", channel_id="123"
    )
    connector = DiscordConnector("test_discord", config)
    connector.state_file = tmp_path / "discord_state.json"

    # Mock API fetch response
    mock_response = MagicMock()
    mock_response.__enter__.return_value = mock_response
    mock_response.read.return_value = json.dumps([
        {"id": "2", "content": "i am bot", "author": {"bot": True}}
    ]).encode("utf-8")
    
    mock_urlopen.side_effect = [mock_response]

    connector.poll_and_reply()

    # Chat logic should not be triggered, but state should skip to msg ID 2
    assert connector.state_file.exists()
    state = json.loads(connector.state_file.read_text())
    assert state["last_message_id"] == "2"


# === Telegram Connector Tests ===

@patch("kage.connectors.telegram.urllib.request.urlopen")
@patch("kage.connectors.telegram.generate_chat_reply")
def test_telegram_connector_poll_and_reply(mock_chat, mock_urlopen, tmp_path):
    import time
    config = TelegramConnectorConfig(
        poll=True, bot_token="123456:ABC", chat_id="789"
    )
    connector = TelegramConnector("test_telegram", config)
    connector.state_file = tmp_path / "telegram_state.json"
    connector.history_file = tmp_path / "telegram_history.jsonl"

    now_unix = int(time.time())

    # Mock API: getUpdates
    mock_updates_response = MagicMock()
    mock_updates_response.__enter__.return_value = mock_updates_response
    mock_updates_response.read.return_value = json.dumps({
        "ok": True,
        "result": [
            {
                "update_id": 100,
                "message": {
                    "message_id": 1,
                    "from": {"id": 42, "is_bot": False, "first_name": "User"},
                    "chat": {"id": 789},
                    "date": now_unix,
                    "text": "hello kage"
                }
            }
        ]
    }).encode("utf-8")

    # Mock API: getMe
    mock_me_response = MagicMock()
    mock_me_response.__enter__.return_value = mock_me_response
    mock_me_response.read.return_value = json.dumps({
        "ok": True,
        "result": {"id": 999, "is_bot": True, "username": "kage_bot"}
    }).encode("utf-8")

    # Mock API: sendMessage
    mock_send_response = MagicMock()
    mock_send_response.__enter__.return_value = mock_send_response
    mock_send_response.read.return_value = json.dumps({
        "ok": True,
        "result": {"message_id": 2}
    }).encode("utf-8")

    # urlopen calls: getUpdates, getMe, sendMessage
    mock_urlopen.side_effect = [mock_updates_response, mock_me_response, mock_send_response]

    # Mock AI reply
    mock_chat.return_value = {"stdout": "hello human", "stderr": "", "returncode": 0}

    connector.poll_and_reply()

    # Check state file updated
    assert connector.state_file.exists()
    state = json.loads(connector.state_file.read_text())
    assert state["last_update_id"] == "100"

    # AI Chat triggered correctly
    mock_chat.assert_called_once()
    actual_prompt = mock_chat.call_args[0][0]
    assert "[Recent Chat History]" in actual_prompt
    assert "[Current Instruction]\nhello kage" in actual_prompt


@patch("kage.connectors.telegram.urllib.request.urlopen")
def test_telegram_connector_ignores_bot_messages(mock_urlopen, tmp_path):
    import time
    config = TelegramConnectorConfig(
        poll=True, bot_token="123456:ABC", chat_id="789"
    )
    connector = TelegramConnector("test_telegram", config)
    connector.state_file = tmp_path / "telegram_state.json"
    connector.history_file = tmp_path / "telegram_history.jsonl"

    now_unix = int(time.time())

    # Mock API: getUpdates — bot message only
    mock_updates_response = MagicMock()
    mock_updates_response.__enter__.return_value = mock_updates_response
    mock_updates_response.read.return_value = json.dumps({
        "ok": True,
        "result": [
            {
                "update_id": 200,
                "message": {
                    "message_id": 5,
                    "from": {"id": 999, "is_bot": True, "first_name": "Kage Bot"},
                    "chat": {"id": 789},
                    "date": now_unix,
                    "text": "i am bot"
                }
            }
        ]
    }).encode("utf-8")

    mock_urlopen.side_effect = [mock_updates_response]

    connector.poll_and_reply()

    # State should advance to update_id 200 but no AI call should be made
    assert connector.state_file.exists()
    state = json.loads(connector.state_file.read_text())
    assert state["last_update_id"] == "200"



import sqlite3
import uuid
from datetime import datetime
from .config import KAGE_DB_PATH


def init_db():
    KAGE_DB_PATH.parent.mkdir(parents=True, exist_ok=True)
    conn = sqlite3.connect(KAGE_DB_PATH)
    cursor = conn.cursor()
    cursor.execute("""
        CREATE TABLE IF NOT EXISTS executions (
            id TEXT PRIMARY KEY,
            project_path TEXT,
            task_name TEXT,
            run_at TEXT,
            status TEXT,
            stdout TEXT,
            stderr TEXT,
            pid INTEGER
        )
    """)
    # Migration: finished_at カラムを追加（既存DB対応）
    try:
        cursor.execute("ALTER TABLE executions ADD COLUMN finished_at TEXT")
    except sqlite3.OperationalError:
        pass  # 既に存在する場合

    # Migration: pid カラムを追加（既存DB対応）
    try:
        cursor.execute("ALTER TABLE executions ADD COLUMN pid INTEGER")
    except sqlite3.OperationalError:
        pass

    conn.commit()
    conn.close()


def start_execution(project_path: str, task_name: str, pid: int = None) -> str:
    """実行開始を記録し、実行IDを返す。"""
    conn = sqlite3.connect(KAGE_DB_PATH)
    cursor = conn.cursor()
    run_at = datetime.now().isoformat()
    exec_id = str(uuid.uuid4())
    cursor.execute(
        """
        INSERT INTO executions (id, project_path, task_name, run_at, status, stdout, stderr, finished_at, pid)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)
    """,
        (exec_id, project_path, task_name, run_at, "RUNNING", "", "", None, pid),
    )
    conn.commit()
    conn.close()
    return exec_id


def get_execution_pid(exec_id: str) -> int | None:
    """実行IDからPIDを取得する。"""
    conn = sqlite3.connect(KAGE_DB_PATH)
    cursor = conn.cursor()
    cursor.execute("SELECT pid FROM executions WHERE id = ?", (exec_id,))
    row = cursor.fetchone()
    conn.close()
    return row[0] if row else None


def update_execution(
    exec_id: str, status: str, stdout: str, stderr: str, finished_at: str = None
):
    """実行結果を更新する。"""
    conn = sqlite3.connect(KAGE_DB_PATH)
    cursor = conn.cursor()
    if finished_at is None:
        finished_at = datetime.now().isoformat()

    cursor.execute(
        """
        UPDATE executions 
        SET status = ?, stdout = ?, stderr = ?, finished_at = ?
        WHERE id = ?
    """,
        (status, stdout, stderr, finished_at, exec_id),
    )

    # ログを最大100件に制限
    cursor.execute("""
        DELETE FROM executions 
        WHERE id NOT IN (
            SELECT id FROM executions 
            ORDER BY run_at DESC 
            LIMIT 100
        )
    """)

    conn.commit()
    conn.close()


def log_execution(
    project_path: str, task_name: str, status: str, stdout: str, stderr: str
):
    """(互換性のために残す) 即時に完了した実行を記録する。"""
    exec_id = start_execution(project_path, task_name)
    update_execution(exec_id, status, stdout, stderr)



import json
import urllib.request
import urllib.error
from pathlib import Path
from datetime import datetime, timezone
from ..ai.chat import generate_chat_reply, clean_ai_reply
from .base import BaseConnector

class DiscordConnector(BaseConnector):
    def __init__(self, name: str, config):
        super().__init__(name, config)
        self.state_file = Path.home() / ".kage" / "connectors" / f"{self.name}_state.json"
        self.history_file = Path.home() / ".kage" / "connectors" / f"{self.name}_history.jsonl"
        
    def _load_state(self):
        if self.state_file.exists():
            try:
                return json.loads(self.state_file.read_text())
            except Exception:
                return {}
        return {}

    def _save_state(self, state):
        self.state_file.parent.mkdir(parents=True, exist_ok=True)
        self.state_file.write_text(json.dumps(state, ensure_ascii=False))

    def _log_history(self, role: str, content: str):
        if not content:
            return
        self.history_file.parent.mkdir(parents=True, exist_ok=True)
        import time
        entry = {"timestamp": int(time.time()), "role": role, "content": content}
        with self.history_file.open("a", encoding="utf-8") as f:
            f.write(json.dumps(entry, ensure_ascii=False) + "\n")

    def _get_bot_identity(self):
        """Fetch bot's own user id and username using /users/@me."""
        url = "https://discord.com/api/v10/users/@me"
        req = urllib.request.Request(url, headers={
            "Authorization": f"Bot {self.config.bot_token}",
            "User-Agent": "kage-connector"
        })
        try:
            with urllib.request.urlopen(req) as response:
                data = json.loads(response.read().decode())
                return data.get("id"), data.get("username")
        except Exception:
            pass
        return None, None

    def poll_and_reply(self):
        if not self.config.bot_token or not self.config.channel_id:
            return

        state = self._load_state()
        last_message_id = state.get("last_message_id")

        limit = max(1, min(100, self.config.history_limit))
        url = f"https://discord.com/api/v10/channels/{self.config.channel_id}/messages?limit={limit}"

        req = urllib.request.Request(url, headers={
            "Authorization": f"Bot {self.config.bot_token}",
            "User-Agent": "kage-connector"
        })

        try:
            with urllib.request.urlopen(req) as response:
                messages = json.loads(response.read().decode())
        except urllib.error.URLError:
            return
        except Exception:
            return

        if not messages:
            return

        # Fetch our own identity to help the AI recognize itself
        bot_user_id, bot_name = self._get_bot_identity()
        identity_context = ""
        if bot_user_id:
            identity_context = f"[YOUR IDENTITY ON DISCORD]\n- Your Username: {bot_name}\n- Your User ID: {bot_user_id}\n- Mentions like <@{bot_user_id}> or <@!{bot_user_id}> refer to YOU. Do not greet yourself.\n\n"

        # Discord API returns newer messages first. Sort oldest to newest.
        messages.sort(key=lambda x: int(x["id"]))

        # If the very last (newest) message in the channel is from a bot,
        # it means the bot already replied, so we shouldn't process anything.
        if messages and messages[-1].get("author", {}).get("bot"):
            # Still update state so we don't process these again
            state["last_message_id"] = messages[-1]["id"]
            self._save_state(state)
            return

        # Build recent message history string to inject as context
        history_lines = []
        for msg in messages:
            author_name = msg.get("author", {}).get("username", "Unknown")
            is_bot = msg.get("author", {}).get("bot", False)
            content_text = msg.get("content", "").strip()
            if content_text:
                role = "Assistant" if is_bot else author_name
                if bot_user_id and msg.get("author", {}).get("id") == bot_user_id:
                    role = "Assistant"
                history_lines.append(f"{role}: {content_text}")
        history_context = "\n".join(history_lines)

        # Find the single newest user message to respond to.
        # We only respond to ONE message per poll to avoid long processing
        # during which another cron cycle could start and cause duplicates.
        target_msg = None
        newest_id = last_message_id

        for msg in messages:
            msg_id = msg["id"]

            # Skip messages we've already processed
            if last_message_id and int(msg_id) <= int(last_message_id):
                continue

            # Always advance the watermark so we don't re-scan these
            newest_id = msg_id

            if msg.get("author", {}).get("bot"):
                continue

            # Filtering by user_id
            if self.config.user_id:
                author_id = msg.get("author", {}).get("id")
                if author_id != str(self.config.user_id):
                    continue

            content = msg.get("content", "").strip()
            if not content:
                continue

            # Filtering by message age
            try:
                msg_time = datetime.fromisoformat(msg["timestamp"])
                age = (datetime.now(timezone.utc) - msg_time).total_seconds()
                if age > self.config.max_age_seconds:
                    continue
            except Exception:
                continue  # If timestamp can't be parsed, skip (don't process stale msgs)

            # This is a valid candidate — keep the NEWEST one
            target_msg = msg

        # Always advance watermark to newest seen message, even if we don't reply
        if newest_id and newest_id != last_message_id:
            state["last_message_id"] = newest_id
            self._save_state(state)

        # Process the single target message (if any)
        if target_msg:
            content = target_msg.get("content", "").strip()
            try:
                prompt_with_history = f"{identity_context}[Recent Chat History]\n{history_context}\n\n[Current Instruction]\n{content}"
                self._log_history("User", content)
                reply_data = generate_chat_reply(
                    prompt_with_history, 
                    system_prompt=self.config.system_prompt,
                    working_dir=self.config.working_dir
                )
                reply_text = reply_data.get("stdout", "")
            except Exception as e:
                reply_text = f"Error generating reply: {e}"

            final_reply_text = clean_ai_reply(reply_text)
            last_reply_id = self._post_reply(final_reply_text)

            # Advance past bot's own reply to prevent self-reply on next poll
            if last_reply_id:
                state["last_message_id"] = last_reply_id
                self._save_state(state)

    def send_message(self, text: str):
        if not self.config.bot_token or not self.config.channel_id:
            return
        self._post_reply(clean_ai_reply(text))

    def _post_reply(self, text) -> str | None:
        """Post a reply, splitting if needed. Returns the last posted message ID."""
        if not text:
            return None
            
        self._log_history("Assistant", text)
        
        # Split into chunks respecting Discord's 2000 char limit
        chunks = self._split_message(text, max_len=1950)
        last_id = None
        for chunk in chunks:
            msg_id = self._send_chunk(chunk)
            if msg_id:
                last_id = msg_id
        return last_id

    @staticmethod
    def _split_message(text: str, max_len: int = 1950) -> list[str]:
        """Split a message into chunks that fit within Discord's character limit.
        Tries to split at newline boundaries when possible."""
        if len(text) <= max_len:
            return [text]
        
        chunks = []
        remaining = text
        while remaining:
            if len(remaining) <= max_len:
                chunks.append(remaining)
                break
            
            # Try to find a newline to split at
            split_at = remaining.rfind("\n", 0, max_len)
            if split_at <= 0:
                # No good newline found, split at max_len
                split_at = max_len
            
            chunks.append(remaining[:split_at])
            remaining = remaining[split_at:].lstrip("\n")
        
        return chunks

    def _send_chunk(self, text: str) -> str | None:
        """Send a single message chunk to Discord. Returns the posted message ID."""
        url = f"https://discord.com/api/v10/channels/{self.config.channel_id}/messages"
        payload = {"content": text}

        data = json.dumps(payload).encode()
        req = urllib.request.Request(url, data=data, headers={
            "Authorization": f"Bot {self.config.bot_token}",
            "Content-Type": "application/json",
            "User-Agent": "kage-connector"
        }, method="POST")

        try:
            with urllib.request.urlopen(req) as response:
                res_data = json.loads(response.read().decode())
                return res_data.get("id")
        except Exception:
            return None




import json
import urllib.request
import urllib.error
from pathlib import Path
from datetime import datetime, timezone
from ..ai.chat import generate_chat_reply, clean_ai_reply
from .base import BaseConnector

class TelegramConnector(BaseConnector):
    def __init__(self, name: str, config):
        super().__init__(name, config)
        self.state_file = Path.home() / ".kage" / "connectors" / f"{self.name}_state.json"
        self.history_file = Path.home() / ".kage" / "connectors" / f"{self.name}_history.jsonl"
        self.api_base = f"https://api.telegram.org/bot{self.config.bot_token}"
        
    def _load_state(self):
        if self.state_file.exists():
            try:
                return json.loads(self.state_file.read_text())
            except Exception:
                return {}
        return {}

    def _save_state(self, state):
        self.state_file.parent.mkdir(parents=True, exist_ok=True)
        self.state_file.write_text(json.dumps(state, ensure_ascii=False))

    def _log_history(self, role: str, content: str):
        if not content:
            return
        self.history_file.parent.mkdir(parents=True, exist_ok=True)
        import time
        entry = {"timestamp": int(time.time()), "role": role, "content": content}
        with self.history_file.open("a", encoding="utf-8") as f:
            f.write(json.dumps(entry, ensure_ascii=False) + "\n")

    def _get_bot_identity(self):
        """Fetch bot's own user id and username using getMe."""
        url = f"{self.api_base}/getMe"
        req = urllib.request.Request(url, headers={
            "User-Agent": "kage-connector"
        })
        try:
            with urllib.request.urlopen(req) as response:
                data = json.loads(response.read().decode())
                if data.get("ok"):
                    result = data["result"]
                    return str(result.get("id")), result.get("username")
        except Exception:
            pass
        return None, None

    def poll_and_reply(self):
        if not self.config.bot_token or not self.config.chat_id:
            return

        state = self._load_state()
        last_update_id = state.get("last_update_id")

        # Telegram getUpdates: fetch new updates since last_update_id
        limit = max(1, min(100, self.config.history_limit))
        url = f"{self.api_base}/getUpdates?limit={limit}&timeout=0"
        if last_update_id:
            url += f"&offset={int(last_update_id) + 1}"

        req = urllib.request.Request(url, headers={
            "User-Agent": "kage-connector"
        })

        try:
            with urllib.request.urlopen(req) as response:
                data = json.loads(response.read().decode())
                if not data.get("ok"):
                    return
                updates = data.get("result", [])
        except urllib.error.URLError:
            return
        except Exception:
            return

        if not updates:
            return

        # Fetch our own identity
        bot_user_id, bot_name = self._get_bot_identity()
        identity_context = ""
        if bot_user_id:
            identity_context = f"[YOUR IDENTITY ON TELEGRAM]\n- Your Username: @{bot_name}\n- Your User ID: {bot_user_id}\n- Do not greet yourself.\n\n"

        # Filter updates that are text messages in the target chat
        chat_messages = []
        newest_update_id = last_update_id

        for update in updates:
            update_id = update.get("update_id")
            if update_id:
                newest_update_id = str(update_id)

            msg = update.get("message")
            if not msg:
                continue

            # Only process messages from the configured chat
            msg_chat_id = str(msg.get("chat", {}).get("id", ""))
            if msg_chat_id != str(self.config.chat_id):
                continue

            chat_messages.append(msg)

        # Always advance watermark
        if newest_update_id and newest_update_id != last_update_id:
            state["last_update_id"] = newest_update_id
            self._save_state(state)

        if not chat_messages:
            return

        # Build recent message history
        history_lines = []
        for msg in chat_messages:
            from_user = msg.get("from", {})
            is_bot = from_user.get("is_bot", False)
            content_text = msg.get("text", "").strip()
            if content_text:
                if is_bot and bot_user_id and str(from_user.get("id")) == bot_user_id:
                    role = "Assistant"
                elif is_bot:
                    role = "Bot"
                else:
                    first = from_user.get("first_name", "")
                    last = from_user.get("last_name", "")
                    role = f"{first} {last}".strip() or from_user.get("username", "User")
                history_lines.append(f"{role}: {content_text}")
        history_context = "\n".join(history_lines)

        # Find the single newest user message to respond to
        target_msg = None

        for msg in chat_messages:
            from_user = msg.get("from", {})

            # Skip bot messages
            if from_user.get("is_bot", False):
                continue

            # Filtering by user_id
            if self.config.user_id:
                author_id = str(from_user.get("id", ""))
                if author_id != str(self.config.user_id):
                    continue

            content = msg.get("text", "").strip()
            if not content:
                continue

            # Filtering by message age
            try:
                msg_date = msg.get("date", 0)
                age = datetime.now(timezone.utc).timestamp() - msg_date
                if age > self.config.max_age_seconds:
                    continue
            except Exception:
                continue

            # Keep the NEWEST one
            target_msg = msg

        # Process the single target message (if any)
        if target_msg:
            content = target_msg.get("text", "").strip()
            try:
                prompt_with_history = f"{identity_context}[Recent Chat History]\n{history_context}\n\n[Current Instruction]\n{content}"
                self._log_history("User", content)
                reply_data = generate_chat_reply(
                    prompt_with_history, 
                    system_prompt=self.config.system_prompt,
                    working_dir=self.config.working_dir
                )
                reply_text = reply_data.get("stdout", "")
            except Exception as e:
                reply_text = f"Error generating reply: {e}"

            final_reply_text = clean_ai_reply(reply_text)
            self._post_reply(final_reply_text)

    def send_message(self, text: str):
        if not self.config.bot_token or not self.config.chat_id:
            return
        self._post_reply(clean_ai_reply(text))

    def _post_reply(self, text) -> str | None:
        """Post a reply, splitting if needed. Returns the last posted message ID."""
        if not text:
            return None
            
        self._log_history("Assistant", text)
        
        # Split into chunks respecting Telegram's 4096 char limit
        chunks = self._split_message(text, max_len=4000)
        last_id = None
        for chunk in chunks:
            msg_id = self._send_chunk(chunk)
            if msg_id:
                last_id = msg_id
        return last_id

    @staticmethod
    def _split_message(text: str, max_len: int = 4000) -> list[str]:
        """Split a message into chunks that fit within Telegram's character limit."""
        if len(text) <= max_len:
            return [text]
        
        chunks = []
        remaining = text
        while remaining:
            if len(remaining) <= max_len:
                chunks.append(remaining)
                break
            
            split_at = remaining.rfind("\n", 0, max_len)
            if split_at <= 0:
                split_at = max_len
            
            chunks.append(remaining[:split_at])
            remaining = remaining[split_at:].lstrip("\n")
        
        return chunks

    def _send_chunk(self, text: str) -> str | None:
        """Send a single message chunk to Telegram. Returns the posted message ID."""
        url = f"{self.api_base}/sendMessage"
        payload = {
            "chat_id": self.config.chat_id,
            "text": text
        }

        data = json.dumps(payload).encode()
        req = urllib.request.Request(url, data=data, headers={
            "Content-Type": "application/json",
            "User-Agent": "kage-connector"
        }, method="POST")

        try:
            with urllib.request.urlopen(req) as response:
                res_data = json.loads(response.read().decode())
                if res_data.get("ok"):
                    return str(res_data.get("result", {}).get("message_id"))
        except Exception:
            pass
        return None



from datetime import datetime, timezone as dt_timezone
import zoneinfo
from croniter import croniter
from pathlib import Path
from .config import KAGE_PROJECTS_LIST, get_global_config
from .parser import load_project_tasks, TaskDef
from .executor import execute_task


def get_projects() -> list[Path]:
    if not KAGE_PROJECTS_LIST.exists():
        return []
    with open(KAGE_PROJECTS_LIST, "r", encoding="utf-8") as f:
        lines = [line.strip() for line in f if line.strip()]
    return [Path(p) for p in set(lines) if Path(p).exists()]


def should_run(cron_expr: str, now: datetime, tz_name: str = "UTC") -> bool:
    try:
        tz = zoneinfo.ZoneInfo(tz_name)
    except zoneinfo.ZoneInfoNotFoundError:
        print(f"Invalid timezone: {tz_name}, falling back to UTC")
        tz = dt_timezone.utc

    # `now` を指定タイムゾーンに変換（awareなdatetimeにする）
    if now.tzinfo is None:
        now_aware = now.astimezone(tz)
    else:
        now_aware = now.astimezone(tz)

    try:
        # croniterにawareなdatetimeを渡す
        itr = croniter(cron_expr, now_aware)
        # crontabは0秒起動が基本なので、nowの秒を切り捨てて判定する
        now_floored = now_aware.replace(second=0, microsecond=0)

        # get_prev で直前の実行予定時刻を取得（これも aware datetime になる）
        prev = itr.get_prev(datetime)

        # 前回の予定実行時刻とnow_flooredが一致すれば実行
        return prev == now_floored
    except Exception as e:
        print(f"Invalid cron expression: {cron_expr}, Error: {e}")
        return False


def parse_hour_string(s: str) -> set[int]:
    """
    "9-17,21" のような文字列を {9,10,11,12,13,14,15,16,17,21} のセットに変換する。
    """
    hours = set()
    if not s:
        return hours

    parts = s.split(",")
    for part in parts:
        part = part.strip()
        if "-" in part:
            try:
                start_s, end_s = part.split("-")
                start, end = int(start_s), int(end_s)
                # 逆転指定 (22-4) はサポートせず、順序通りとする
                for h in range(start, end + 1):
                    if 0 <= h <= 23:
                        hours.add(h)
            except ValueError:
                continue
        else:
            try:
                h = int(part)
                if 0 <= h <= 23:
                    hours.add(h)
            except ValueError:
                continue
    return hours


def is_within_time_window(task: TaskDef, now: datetime, tz_name: str) -> bool:
    try:
        tz = zoneinfo.ZoneInfo(tz_name)
    except Exception:
        tz = dt_timezone.utc

    local_now = now.astimezone(tz)
    current_hour = local_now.hour

    # 禁止時間帯のチェック
    if task.denied_hours:
        denied = parse_hour_string(task.denied_hours)
        if current_hour in denied:
            return False

    # 許可時間帯のチェック
    if task.allowed_hours:
        allowed = parse_hour_string(task.allowed_hours)
        if current_hour not in allowed:
            return False

    return True


def run_all_scheduled_tasks():
    # utcnow を基準にしてタイムゾーン変換に備える
    now = datetime.now(dt_timezone.utc)
    projects = get_projects()
    cfg = get_global_config()
    tz_name = cfg.timezone

    for proj_dir in projects:
        tasks = load_project_tasks(proj_dir)
        for toml_file, local_task in tasks:
            t = local_task.task
            if not t.active:
                continue

            # タスク固有のタイムゾーン、なければグローバル
            task_tz = t.timezone or tz_name

            # 時間帯枠のチェック
            if not is_within_time_window(t, now, task_tz):
                continue

            if should_run(t.cron, now, task_tz):
                # 実行条件を満たした場合、非同期またはバックグラウンドで実行
                # 今回はPython内のサブプロセスとして呼び出す
                execute_task(proj_dir, local_task.task, task_file=toml_file)
    
    # Run active connectors
    from .connectors.runner import run_connectors
    import threading
    
    # Run connectors in a separate thread so it doesn't block if there are many tasks, 
    # but since this is at the end, we can just call it synchronously or run it before tasks 
    # in a background thread that we join at the end.
    pass  # We will just run it at the end for simplicity.

    run_connectors()



import typer
from . import config as config_mod, daemon, db
from typing import Optional
from importlib import metadata
from pathlib import Path
from typer import completion as typer_completion

app = typer.Typer(
    help="kage - AI Native Cron Task Runner",
    add_completion=True,
)

cron_app = typer.Typer(help="OS-level scheduler (cron/launchd) management")
app.add_typer(cron_app, name="cron")

task_app = typer.Typer(help="Manage kage tasks")
app.add_typer(task_app, name="task")

project_app = typer.Typer(help="Manage registered projects")
app.add_typer(project_app, name="project")

connector_app = typer.Typer(help="Manage chat connectors (Discord, Slack, Telegram, etc.)")
app.add_typer(connector_app, name="connector")

completion_app = typer.Typer(help="Shell completion helpers")
app.add_typer(completion_app, name="completion")


def _completion_script(shell: str) -> str:
    target_shell = shell.lower().strip()
    if target_shell not in ("bash", "zsh"):
        raise typer.BadParameter("shell must be one of: bash, zsh")
    return typer_completion.get_completion_script(
        prog_name="kage",
        complete_var="_KAGE_COMPLETE",
        shell=target_shell,
    )


def _append_source_line_if_missing(rc_file: Path, source_line: str) -> bool:
    rc_file.parent.mkdir(parents=True, exist_ok=True)
    if not rc_file.exists():
        rc_file.write_text("", encoding="utf-8")
    content = rc_file.read_text(encoding="utf-8")
    if source_line in content:
        return False
    with rc_file.open("a", encoding="utf-8") as f:
        if content and not content.endswith("\n"):
            f.write("\n")
        f.write(source_line + "\n")
    return True


@completion_app.command("show")
def completion_show(
    shell: str = typer.Argument(..., help="Target shell: bash or zsh"),
):
    """Print completion script for bash/zsh."""
    typer.echo(_completion_script(shell))


@completion_app.command("install")
def completion_install(
    shell: str = typer.Argument(..., help="Target shell: bash or zsh"),
):
    """Install completion script and update shell rc file."""
    target_shell = shell.lower().strip()
    script = _completion_script(target_shell)

    comp_dir = Path.home() / ".kage" / "completions"
    comp_dir.mkdir(parents=True, exist_ok=True)
    script_path = comp_dir / f"kage.{target_shell}"
    script_path.write_text(script, encoding="utf-8")

    if target_shell == "bash":
        rc_file = Path.home() / ".bashrc"
    else:
        rc_file = Path.home() / ".zshrc"
    source_line = f'source "{script_path}"'
    appended = _append_source_line_if_missing(rc_file, source_line)

    typer.echo(f"Installed completion script: {script_path}")
    if appended:
        typer.echo(f"Updated shell config: {rc_file}")
    else:
        typer.echo(f"Shell config already contains source line: {rc_file}")
    typer.echo("Reload your shell: exec $SHELL -l")


def _resolve_version() -> str:
    for pkg in ("kage-ai", "kage"):
        try:
            return metadata.version(pkg)
        except metadata.PackageNotFoundError:
            continue
    return "unknown"


def _version_callback(value: bool):
    if value:
        typer.echo(_resolve_version())
        raise typer.Exit()


@app.callback()
def app_callback(
    version: bool = typer.Option(
        False,
        "--version",
        "-V",
        help="Show version and exit.",
        is_eager=True,
        callback=_version_callback,
    ),
):
    """kage CLI."""
    return None


@project_app.command("list")
def project_list():
    """List all registered projects."""
    from .scheduler import get_projects
    from .parser import load_project_tasks
    from .config import KAGE_PROJECTS_LIST
    from rich.console import Console
    from rich.table import Table

    console = Console()
    projects = get_projects()

    if not projects:
        console.print("[yellow]No projects registered.[/yellow]")
        console.print(
            "Run [bold]kage init[/bold] in a project directory to register it."
        )
        console.print(f"List file: {KAGE_PROJECTS_LIST}")
        return

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Project Path", style="bold")
    table.add_column("Tasks", justify="right")
    table.add_column("Status")

    for proj_dir in sorted(projects):
        tasks = load_project_tasks(proj_dir)
        task_count = len(tasks)
        exists = proj_dir.exists()
        status = "[green]✔ OK[/green]" if exists else "[red]✘ Directory not found[/red]"
        table.add_row(str(proj_dir), str(task_count), status)

    console.print(table)
    console.print(f"[dim]Source: {KAGE_PROJECTS_LIST}[/dim]")


@project_app.command("remove")
def project_remove(
    path: Optional[str] = typer.Argument(
        None, help="Project path to unregister (defaults to current directory)"
    ),
):
    """Unregister a project from kage."""
    from .config import KAGE_PROJECTS_LIST
    from rich.console import Console
    from pathlib import Path

    console = Console()
    if path:
        target_path = Path(path).resolve()
    else:
        target_path = Path.cwd().resolve()
    target = str(target_path)

    if not KAGE_PROJECTS_LIST.exists():
        console.print("[yellow]No projects registered.[/yellow]")
        return

    lines = KAGE_PROJECTS_LIST.read_text().splitlines()
    new_lines = [
        line
        for line in lines
        if line.strip() and str(Path(line.strip()).resolve()) != target
    ]

    if len(new_lines) == len([line for line in lines if line.strip()]):
        console.print(f"[yellow]Project not found in registry: {target}[/yellow]")
        return

    KAGE_PROJECTS_LIST.write_text("\n".join(new_lines) + ("\n" if new_lines else ""))
    console.print(f"[green]✔ Removed:[/green] {target}")


@task_app.command("list")
def task_list(
    project: Optional[str] = typer.Option(
        None, "--project", "-p", help="Filter by project path"
    ),
):
    """List all registered tasks and their status (ON/OFF)."""
    from .scheduler import get_projects
    from .parser import load_project_tasks
    from rich.console import Console
    from rich.table import Table

    console = Console()
    projects = get_projects()

    if not projects:
        console.print(
            "[yellow]No projects registered. Run 'kage init' in a project directory.[/yellow]"
        )
        raise typer.Exit()

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Project", style="dim")
    table.add_column("Task Name", style="bold")
    table.add_column("Active")
    table.add_column("Schedule")
    table.add_column("Type")
    table.add_column("Provider/Command")

    found = False
    for proj_dir in projects:
        if project and project not in str(proj_dir):
            continue
        tasks = load_project_tasks(proj_dir)
        for toml_file, local_task in tasks:
            t = local_task.task
            task_type = "AI Prompt" if t.prompt else "Shell"
            provider_info = (
                t.provider or (t.ai.engine if t.ai and t.ai.engine else "")
                if t.prompt
                else (t.command or "")[:40]
            )
            status = "[green]ON[/green]" if t.active else "[red]OFF[/red]"
            table.add_row(
                str(proj_dir),
                t.name,
                status,
                t.cron,
                task_type,
                provider_info or "-",
            )
            found = True

    if not found:
        console.print("[yellow]No tasks found.[/yellow]")
    else:
        console.print(table)


@task_app.command("new")
def task_new(
    file_name: str = typer.Argument(
        ..., help="Name of the task file (without extension)"
    ),
):
    """Create a new kage task file in .kage/tasks/."""
    from pathlib import Path
    from rich.console import Console

    console = Console()
    tasks_dir = Path.cwd() / ".kage" / "tasks"

    if not tasks_dir.parent.exists():
        console.print("[red]Not in a kage project. Run 'kage init' first.[/red]")
        raise typer.Exit(1)

    tasks_dir.mkdir(parents=True, exist_ok=True)
    target_path = tasks_dir / f"{file_name}.md"

    if target_path.exists():
        console.print(f"[yellow]Task file already exists: {target_path}[/yellow]")
        raise typer.Exit(1)

    import locale
    lang = "en"
    try:
        loc, _ = locale.getlocale()
        if loc and loc.startswith("ja"):
            lang = "ja"
    except Exception:
        pass
    
    import os
    if os.environ.get("LANG", "").startswith("ja"):
        lang = "ja"

    if lang == "ja":
        template = f"""---
name: {file_name.replace("_", " ").title()}
cron: "0 3 * * *"
active: false
mode: autostop
---

# Task: PDFのOCR精度測定ベンチマーク

PDFからテキストを抽出する最適な無料OCRモデルを選定するため、ベンチマークテストを実施してください。一晩かけて1つずつモデルを検証し、最終的な比較レポートを作成してください。

1. **データ準備**: サンプルPDFが存在しない場合は、テスト用に適当な日本語のダミーPDFを作成（またはダウンロード）してください。
2. **モデル検証**: 以下のOCRツールを1回の実行（run）につき1つずつインストール・実行し、テキスト抽出の精度と処理速度を計測してください。
   - Tesseract OCR (with jpn data)
   - EasyOCR
   - PaddleOCR
   - marker (Surya)
3. **レポート作成**: すべての検証が完了したら、`ocr_benchmark_report.md` をルートディレクトリに作成し、各モデルの精度、速度、導入のしやすさなどを比較した表を出力してください。
4. **終了**: レポートが出力されたら、すべてのサブタスクを 'done' にしてこのタスクを停止してください。
"""
    else:
        template = f"""---
name: {file_name.replace("_", " ").title()}
cron: "0 3 * * *"
active: false
mode: autostop
---

# Task: PDF OCR Accuracy Benchmark

We need to select the best free OCR model for extracting text from PDFs. Please conduct a benchmark test overnight, evaluating one model per run, and create a final comparison report.

1. **Data Prep**: If a sample PDF doesn't exist, create (or download) a dummy PDF with varied text layouts for testing.
2. **Model Evaluation**: Install and run one of the following OCR tools per execution run. Measure text extraction accuracy and processing speed.
   - Tesseract OCR
   - EasyOCR
   - PaddleOCR
   - marker (Surya)
3. **Reporting**: Once all evaluations are complete, generate `ocr_benchmark_report.md` in the root directory. Include a comparison table showing accuracy, speed, and ease of setup for each model.
4. **Completion**: After the report is generated, mark all sub-tasks as 'done' to stop this task automatically.
"""
    target_path.write_text(template, encoding="utf-8")
    console.print(f"[green]✔ Created new task file:[/green] {target_path}")


def _set_task_active_state(name: Optional[str], active: bool, all_tasks: bool = False):
    from .scheduler import get_projects
    from .parser import load_project_tasks
    from rich.console import Console
    import re

    console = Console()
    projects = get_projects()
    found_any = False

    for proj_dir in projects:
        tasks = load_project_tasks(proj_dir)
        for task_file, local_task in tasks:
            t = local_task.task
            if all_tasks or t.name == name:
                found_any = True
                # ファイルを直接書き換える
                content = task_file.read_text(encoding="utf-8")

                # Markdown の Front Matter を書き換える
                # シンプルに正規表現で active: ... を置換
                new_val = "true" if active else "false"
                if "active:" in content:
                    content = re.sub(
                        r"active:\s*(true|false)", f"active: {new_val}", content
                    )
                else:
                    # active が無い場合は cron: の後に追加
                    content = re.sub(
                        r"(cron:.*?\n)", rf"\1active: {new_val}\n", content
                    )

                task_file.write_text(content, encoding="utf-8")
                state_str = (
                    "[green]ENABLED[/green]" if active else "[red]DISABLED[/red]"
                )
                console.print(f"{state_str}: {t.name} ({task_file})")

    if not found_any and not all_tasks:
        console.print(f"[red]Task '{name}' not found.[/red]")
        raise typer.Exit(1)


@task_app.command("on")
def task_on(
    name: Optional[str] = typer.Argument(None, help="Task name to enable"),
    all_tasks: bool = typer.Option(False, "--all", help="Enable all tasks"),
):
    """Enable a specific task or all tasks."""
    if not name and not all_tasks:
        typer.echo("Error: Must specify a task name or use --all")
        raise typer.Exit(1)
    _set_task_active_state(name, True, all_tasks)


@task_app.command("off")
def task_off(
    name: Optional[str] = typer.Argument(None, help="Task name to disable"),
    all_tasks: bool = typer.Option(False, "--all", help="Disable all tasks"),
):
    """Disable a specific task or all tasks."""
    if not name and not all_tasks:
        typer.echo("Error: Must specify a task name or use --all")
        raise typer.Exit(1)
    _set_task_active_state(name, False, all_tasks)


@task_app.command("show")
def task_show(name: str = typer.Argument(..., help="Task name to show details for")):
    """Show details of a specific task."""
    from .scheduler import get_projects
    from .parser import load_project_tasks
    from rich.console import Console
    from rich.panel import Panel

    console = Console()
    projects = get_projects()

    for proj_dir in projects:
        for toml_file, local_task in load_project_tasks(proj_dir):
            t = local_task.task
            if t.name == name:
                details = [
                    f"[bold]Name:[/bold]           {t.name}",
                    f"[bold]Schedule:[/bold]       {t.cron}",
                    f"[bold]Mode:[/bold]           {t.mode}",
                    f"[bold]Concurrency:[/bold]    {t.concurrency_policy}",
                    f"[bold]Timezone:[/bold]       {t.timezone or 'global'}",
                    f"[bold]Allowed Hours:[/bold]  {t.allowed_hours or 'any'}",
                    f"[bold]Denied Hours:[/bold]   {t.denied_hours or 'none'}",
                    f"[bold]Project:[/bold]        {proj_dir}",
                    f"[bold]File:[/bold]           {toml_file}",
                ]
                if t.prompt:
                    details.append("[bold]Type:[/bold]           AI Prompt")
                    details.append(f"[bold]Prompt:[/bold]         {t.prompt[:100]}...")
                    details.append(
                        f"[bold]Provider:[/bold]       {t.provider or 'global default'}"
                    )
                elif t.command:
                    details.append("[bold]Type:[/bold]           Shell Command")
                    details.append(f"[bold]Command:[/bold]        {t.command}")
                console.print(
                    Panel(
                        "\n".join(details), title=f"[cyan]{t.name}[/cyan]", expand=False
                    )
                )
                return

    console.print(f"[red]Task '{name}' not found.[/red]")
    raise typer.Exit(1)


@task_app.command("run")
def task_run(name: str = typer.Argument(..., help="Task name to run immediately")):
    """Run a specific task immediately (ignores schedule)."""
    from .scheduler import get_projects
    from .parser import load_project_tasks
    from .executor import execute_task
    from rich.console import Console

    console = Console()
    projects = get_projects()

    for proj_dir in projects:
        for toml_file, local_task in load_project_tasks(proj_dir):
            if local_task.task.name == name:
                console.print(
                    f"[cyan]Running task:[/cyan] [bold]{name}[/bold] in {proj_dir}"
                )
                execute_task(proj_dir, local_task.task, task_file=toml_file)
                console.print(f"[green]✓ Task '{name}' completed.[/green]")
                return

    console.print(f"[red]Task '{name}' not found.[/red]")
    raise typer.Exit(1)


@cron_app.command("install")
def cron_install():
    """Register kage run to system scheduler (cron/launchd)."""
    daemon.install()


@cron_app.command("remove")
def cron_remove():
    """Unregister kage run from system scheduler."""
    daemon.remove()


@cron_app.command("status")
def cron_status():
    """Check if kage is registered in system scheduler."""
    daemon.status()


@cron_app.command("start")
def cron_start():
    """Start/Enable background tasks."""
    daemon.start()


@cron_app.command("stop")
def cron_stop():
    """Stop/Disable background tasks."""
    daemon.stop()


@cron_app.command("restart")
def cron_restart():
    """Restart background tasks."""
    daemon.restart()


def _is_ja() -> bool:
    import locale
    import os
    if os.environ.get("LANG", "").startswith("ja"):
        return True
    try:
        loc, _ = locale.getlocale()
        if loc and loc.startswith("ja"):
            return True
    except Exception:
        pass
    return False


@app.command()
def onboard():
    """Initial setup for kage: Create ~/.kage and default configuration."""
    if _is_ja():
        typer.echo("kage の初期セットアップを実行中...")
    else:
        typer.echo("Initializing kage onboard...")
    
    config_mod.setup_global()
    daemon.install()
    db.init_db()
    
    if _is_ja():
        typer.echo("グローバル設定とデータベースのセットアップが完了しました。")
    else:
        typer.echo("Successfully set up global configuration and database.")


@app.command()
def init():
    """Initialize a kage project in the current directory."""
    if _is_ja():
        typer.echo("プロジェクトを初期化中...")
    else:
        typer.echo("Initializing kage project...")
    
    config_mod.setup_local()
    
    if _is_ja():
        typer.echo("初期化が完了しました。")
    else:
        typer.echo("Project initialized.")


@app.command()
def run():
    """Run scheduled tasks."""
    from .scheduler import run_all_scheduled_tasks

    run_all_scheduled_tasks()


@app.command()
def logs(limit: int = 10):
    """View kage execution logs."""
    from .config import KAGE_DB_PATH
    import sqlite3
    from rich.table import Table
    from rich.console import Console

    if not KAGE_DB_PATH.exists():
        typer.echo("No logs found.")
        return

    conn = sqlite3.connect(KAGE_DB_PATH)
    cursor = conn.cursor()
    cursor.execute(
        """
        SELECT run_at, project_path, task_name, status, stdout, stderr, finished_at
        FROM executions 
        ORDER BY run_at DESC LIMIT ?
    """,
        (limit,),
    )
    rows = cursor.fetchall()

    console = Console()
    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Run At")
    table.add_column("Project")
    table.add_column("Task")
    table.add_column("Status")
    table.add_column("Duration", justify="right")
    table.add_column("Output (stdout / stderr)", style="dim")

    for row in rows:
        run_at, proj, name, status, stdout_val, stderr_val, finished_at = row

        # 整形: "2026-02-23T18:28:01.032891" -> "02/23 18:28:01"
        try:
            from datetime import datetime

            dt_start = datetime.fromisoformat(run_at)
            run_at_str = dt_start.strftime("%m/%d %H:%M:%S")
            
            if finished_at:
                dt_end = datetime.fromisoformat(finished_at)
                duration_sec = (dt_end - dt_start).total_seconds()
                duration_str = f"{int(duration_sec)}s"
            else:
                if status == "RUNNING":
                    duration_sec = (datetime.now() - dt_start).total_seconds()
                    duration_str = f"{int(duration_sec)}s+"
                else:
                    duration_str = "-"
        except ValueError:
            run_at_str = run_at[:19].replace("T", " ")
            duration_str = "-"

        # Output 整形
        out_str = []
        if stdout_val and str(stdout_val).strip():
            clean_out = str(stdout_val).strip().replace("\n", " ").replace("\r", "")
            out_str.append(f"[STDOUT] {clean_out[:40]}")
        if stderr_val and str(stderr_val).strip():
            clean_err = str(stderr_val).strip().replace("\n", " ").replace("\r", "")
            out_str.append(f"[STDERR] {clean_err[:40]}")

        final_out = " | ".join(out_str)
        if len(final_out) > 65:
            final_out = final_out[:62] + "..."

        # ステータスの色付け
        if status == "SUCCESS":
            status_colored = f"[green]{status}[/green]"
        elif status == "RUNNING":
            status_colored = f"[yellow]{status}[/yellow]"
        else:
            status_colored = f"[red]{status}[/red]"

        # プロジェクトパスの短縮 (最後の2ディレクトリなどを表示)
        from pathlib import Path

        p = Path(proj)
        proj_str = f".../{p.parent.name}/{p.name}" if len(p.parts) > 2 else proj

        table.add_row(run_at_str, proj_str, str(name), status_colored, duration_str, final_out)

    console.print(table)
    conn.close()


@app.command()
def stop(exec_id: str = typer.Argument(..., help="Execution ID to stop")):
    """Stop a running execution."""
    from .executor import stop_execution
    from rich.console import Console

    console = Console()
    console.print(f"[yellow]Stopping execution {exec_id}...[/yellow]")
    stop_execution(exec_id)
    console.print("[green]Stop signal sent.[/green]")


@app.command()
def ui(
    host: Optional[str] = typer.Option(None, "--host", "-h", help="Bind host (e.g., '0.0.0.0' for external access)"),
    port: Optional[int] = typer.Option(None, "--port", "-p", help="Bind port"),
):
    """Launch the web UI dashboard."""
    from .config import get_global_config
    from .web import start_ui

    cfg = get_global_config()
    target_host = host or cfg.ui_host
    target_port = port or cfg.ui_port
    typer.echo(f"Starting web UI on {target_host}:{target_port}...")
    start_ui(host=target_host, port=target_port)


@app.command()
def config(
    key: str = typer.Argument(..., help="Setting key (e.g., 'default_ai_engine')"),
    value: str = typer.Argument(..., help="New value"),
    is_global: bool = typer.Option(
        False,
        "--global",
        "-g",
        help="Update global config (~/.kage/config.toml) instead of workspace config",
    ),
):
    """Update configuration via CLI."""
    from .config import set_config_value

    set_config_value(key, value, is_global=is_global)


@app.command("config-show")
def config_show(
    workspace: Optional[str] = typer.Option(
        None, "--workspace", "-w", help="Workspace path for resolving .kage/config.toml"
    ),
):
    """Show resolved configuration (commands/providers/defaults)."""
    from pathlib import Path
    from rich.console import Console
    from rich.table import Table
    from .config import get_global_config, KAGE_CONFIG_PATH

    ws_dir = Path(workspace).resolve() if workspace else Path.cwd()
    ws_cfg_path = ws_dir / ".kage" / "config.toml"
    cfg = get_global_config(workspace_dir=ws_dir)

    console = Console()
    console.print("\n[bold cyan]kage config-show[/bold cyan]\n")

    summary = Table(show_header=False, box=None, padding=(0, 1))
    summary.add_column("k", style="bold")
    summary.add_column("v")
    summary.add_row("Workspace", str(ws_dir))
    summary.add_row(
        "Global Config",
        f"{KAGE_CONFIG_PATH} ({'found' if KAGE_CONFIG_PATH.exists() else 'missing'})",
    )
    summary.add_row(
        "Workspace Config",
        f"{ws_cfg_path} ({'found' if ws_cfg_path.exists() else 'missing'})",
    )
    summary.add_row("default_ai_engine", str(cfg.default_ai_engine or "None"))
    summary.add_row("ui_port", str(cfg.ui_port))
    summary.add_row("ui_host", str(cfg.ui_host))
    summary.add_row("log_level", str(cfg.log_level))
    summary.add_row("timezone", str(cfg.timezone))
    summary.add_row("cron_interval_minutes", str(cfg.cron_interval_minutes))
    summary.add_row("env_path", str(cfg.env_path or "None"))
    console.print(summary)

    provider_table = Table(title="Providers", show_header=True, header_style="bold")
    provider_table.add_column("name", style="bold")
    provider_table.add_column("command")
    provider_table.add_column("parser")
    provider_table.add_column("parser_args")
    for name in sorted(cfg.providers.keys()):
        p = cfg.providers[name]
        provider_table.add_row(name, p.command, p.parser, p.parser_args or "")
    if cfg.providers:
        console.print(provider_table)
    else:
        console.print("[yellow]No providers loaded.[/yellow]")

    command_table = Table(title="Commands", show_header=True, header_style="bold")
    command_table.add_column("name", style="bold")
    command_table.add_column("template")
    for name in sorted(cfg.commands.keys()):
        c = cfg.commands[name]
        command_table.add_row(name, " ".join(c.template))
    if cfg.commands:
        console.print(command_table)
    else:
        console.print("[yellow]No commands loaded.[/yellow]")


@app.command()
def doctor():
    """Run setup diagnostics and show potential issues."""
    import os
    from datetime import datetime
    from pathlib import Path
    import tomlkit
    from croniter import croniter
    from rich.console import Console
    from rich.table import Table
    from .config import (
        get_global_config,
        KAGE_GLOBAL_DIR,
        KAGE_CONFIG_PATH,
        KAGE_PROJECTS_LIST,
        KAGE_DB_PATH,
    )

    is_ja = os.environ.get("LANG", "").startswith("ja")

    t_title = "セットアップ診断" if is_ja else "Setup Diagnostics"
    t_chk = "チェック項目" if is_ja else "Check Item"
    t_det = "詳細" if is_ja else "Details"
    t_exists = "存在します" if is_ja else "Found"
    t_run_onboard = "kage onboard を実行してください" if is_ja else "Run 'kage onboard'"
    t_not_found = "が見つかりません" if is_ja else "not found"
    t_only_default = (
        "ライブラリデフォルトのみが使用されます" if is_ja else "Using library defaults"
    )
    t_proj_reg = "プロジェクト登録済み" if is_ja else "projects registered"
    t_run_init = (
        "kage onboard / init を実行してください"
        if is_ja
        else "Run 'kage onboard' or 'kage init'"
    )
    t_ai_not_set = (
        "default_ai_engine が未設定" if is_ja else "default_ai_engine not set"
    )
    t_ai_hint = (
        "AIタスクを使う場合: kage config default_ai_engine codex --global"
        if is_ja
        else "If using AI tasks: run 'kage config default_ai_engine codex --global'"
    )
    t_prov_undef = "が未定義" if is_ja else "is undefined"
    t_prov_cmd_hint = (
        "default_config.toml または config.toml に commands を追加してください"
        if is_ja
        else "Add commands to your config.toml"
    )
    t_prov_hint = (
        "config.toml に providers セクションを追加するか、デフォルト設定を確認してください"
        if is_ja
        else "Check providers in your config or default_config.toml"
    )
    t_res = "結果:" if is_ja else "Result:"
    t_err = "エラー" if is_ja else "errors"
    t_warn = "警告" if is_ja else "warnings"

    console = Console()
    console.print(f"\n[bold cyan]kage doctor[/bold cyan] — {t_title}\n")

    checks = []

    def ok(label, detail=""):
        checks.append(("[green]✔[/green]", label, detail))

    def warn(label, detail=""):
        checks.append(("[yellow]⚠[/yellow]", label, detail))

    def fail(label, detail=""):
        checks.append(("[red]✘[/red]", label, detail))

    def _validate_cron(expr: object, label: str):
        if not isinstance(expr, str):
            fail(label, "cron must be a string")
            return
        try:
            croniter(expr, datetime.utcnow())
        except Exception:
            warn(label, f"invalid cron expression: {expr}")

    def _load_toml(path: Path):
        try:
            with open(path, "r", encoding="utf-8") as f:
                doc = tomlkit.load(f)
            return doc.unwrap() if hasattr(doc, "unwrap") else dict(doc), None
        except Exception as e:
            return None, str(e)

    def _validate_config_file(path: Path, scope: str):
        data, err = _load_toml(path)
        if err:
            fail(f"{scope} parse", err)
            return
        if not isinstance(data, dict):
            fail(scope, "must be a TOML table/object")
            return

        allowed_top = {
            "default_ai_engine",
            "log_level",
            "ui_port",
            "ui_host",
            "cron_interval_minutes",
            "timezone",
            "env_path",
            "commands",
            "providers",
            "connectors",
            "memory_max_entries",
            "working_dir",
        }
        for k in data.keys():
            if k not in allowed_top:
                warn(scope, f"unknown top-level key: {k}")

        typed_keys = {
            "default_ai_engine": (str,),
            "log_level": (str,),
            "ui_port": (int,),
            "ui_host": (str,),
            "cron_interval_minutes": (int,),
            "timezone": (str,),
            "env_path": (str,),
            "system_prompt": (str,),
            "memory_max_entries": (int,),
            "working_dir": (str,),
        }
        for key, expected in typed_keys.items():
            if (
                key in data
                and data[key] is not None
                and not isinstance(data[key], expected)
            ):
                fail(scope, f"{key} must be {expected[0].__name__}")

        commands = data.get("commands")
        if commands is not None:
            if not isinstance(commands, dict):
                fail(scope, "commands must be a table")
            else:
                for name, cmd in commands.items():
                    if not isinstance(cmd, dict):
                        fail(scope, f"commands.{name} must be a table")
                        continue
                    for k in cmd.keys():
                        if k != "template":
                            warn(scope, f"commands.{name}: unknown key {k}")
                    tmpl = cmd.get("template")
                    if not isinstance(tmpl, list) or not all(
                        isinstance(x, str) for x in tmpl
                    ):
                        fail(scope, f"commands.{name}.template must be string array")

        providers = data.get("providers")
        if providers is not None:
            if not isinstance(providers, dict):
                fail(scope, "providers must be a table")
            else:
                for name, prov in providers.items():
                    if not isinstance(prov, dict):
                        fail(scope, f"providers.{name} must be a table")
                        continue
                    for k in prov.keys():
                        if k not in {"command", "parser", "parser_args"}:
                            warn(scope, f"providers.{name}: unknown key {k}")
                    if "command" in prov and not isinstance(prov["command"], str):
                        fail(scope, f"providers.{name}.command must be string")
                    if "parser" in prov and not isinstance(prov["parser"], str):
                        fail(scope, f"providers.{name}.parser must be string")
                    if "parser_args" in prov and not isinstance(
                        prov["parser_args"], str
                    ):
                        fail(scope, f"providers.{name}.parser_args must be string")
                    if isinstance(commands, dict):
                        command_name = prov.get("command")
                        if (
                            isinstance(command_name, str)
                            and command_name not in commands
                        ):
                            warn(
                                scope,
                                f"providers.{name}.command references unknown commands.{command_name}",
                            )

    def _split_markdown_front_matter(text: str):
        lines = text.splitlines()
        if not lines or lines[0].strip() != "---":
            return None, text
        end_idx = None
        for i in range(1, len(lines)):
            if lines[i].strip() == "---":
                end_idx = i
                break
        if end_idx is None:
            return None, text
        data = {}
        for raw in lines[1:end_idx]:
            line = raw.strip()
            if not line or line.startswith("#") or ":" not in line:
                continue
            k, v = line.split(":", 1)
            key = k.strip()
            val = v.strip()
            if (val.startswith('"') and val.endswith('"')) or (
                val.startswith("'") and val.endswith("'")
            ):
                val = val[1:-1]
            data[key] = val
        body = "\n".join(lines[end_idx + 1 :]).strip()
        return data, body

    def _validate_task_mapping(task_data: dict, label: str, merged_cfg):
        allowed = {
            "name",
            "cron",
            "active",
            "mode",
            "concurrency_policy",
            "timezone",
            "timeout_minutes",
            "allowed_hours",
            "denied_hours",
            "command",
            "shell",
            "prompt",
            "provider",
            "command_template",
            "parser",
            "parser_args",
            "ai",
            "connector",
            "connectors",
            "notify_connectors",
        }
        for k in task_data.keys():
            if k not in allowed:
                warn(label, f"unknown key: {k}")

        if (
            not isinstance(task_data.get("name"), str)
            or not task_data.get("name", "").strip()
        ):
            fail(label, "name is required and must be string")
        _validate_cron(task_data.get("cron"), label)

        if task_data.get("prompt") is None and task_data.get("command") is None:
            warn(label, "neither prompt nor command is set")
        if task_data.get("prompt") is not None and not isinstance(
            task_data.get("prompt"), str
        ):
            fail(label, "prompt must be string")
        if task_data.get("command") is not None and not isinstance(
            task_data.get("command"), str
        ):
            fail(label, "command must be string")
        if task_data.get("shell") is not None and not isinstance(
            task_data.get("shell"), str
        ):
            fail(label, "shell must be string")

        command_template = task_data.get("command_template")
        if command_template is not None and (
            not isinstance(command_template, list)
            or not all(isinstance(x, str) for x in command_template)
        ):
            fail(label, "command_template must be string array")

        if task_data.get("provider") is not None:
            if not isinstance(task_data["provider"], str):
                fail(label, "provider must be string")
            elif task_data["provider"] not in merged_cfg.providers:
                warn(
                    label,
                    f"provider '{task_data['provider']}' is not defined in merged config",
                )

        if task_data.get("parser") is not None and not isinstance(
            task_data["parser"], str
        ):
            fail(label, "parser must be string")
        if task_data.get("parser_args") is not None and not isinstance(
            task_data["parser_args"], str
        ):
            fail(label, "parser_args must be string")

        ai = task_data.get("ai")
        if ai is not None:
            if not isinstance(ai, dict):
                fail(label, "ai must be table")
            else:
                for k in ai.keys():
                    if k not in {"engine", "args"}:
                        warn(label, f"ai: unknown key {k}")
                if (
                    "engine" in ai
                    and ai["engine"] is not None
                    and not isinstance(ai["engine"], str)
                ):
                    fail(label, "ai.engine must be string")
                if "args" in ai and ai["args"] is not None:
                    args = ai["args"]
                    if not isinstance(args, list) or not all(
                        isinstance(x, str) for x in args
                    ):
                        fail(label, "ai.args must be string array")

    def _validate_task_file(task_file: Path, merged_cfg):
        label = str(task_file)
        if task_file.suffix.lower() != ".md":
            warn(label, "Only .md task files are supported. TOML tasks are deprecated.")
            return

        try:
            text = task_file.read_text(encoding="utf-8")
        except Exception as e:
            fail(label, f"read error: {e}")
            return

        fm, body = _split_markdown_front_matter(text)
        if not fm:
            fail(label, "missing markdown front matter")
            return

        allowed = {
            "name",
            "cron",
            "active",
            "mode",
            "concurrency_policy",
            "timezone",
            "timeout_minutes",
            "allowed_hours",
            "denied_hours",
            "provider",
            "command",
            "shell",
            "parser",
            "parser_args",
            "connector",
            "connectors",
            "notify_connectors",
        }
        for k in fm.keys():
            if k not in allowed:
                warn(label, f"front matter unknown key: {k}")

        if not isinstance(fm.get("name"), str) or not fm.get("name", "").strip():
            fail(label, "front matter name is required")
        _validate_cron(fm.get("cron"), label)

        if not body.strip() and "command" not in fm:
            fail(
                label,
                "markdown task requires either body prompt or 'command' in front matter",
            )

        task_data = {
            "name": fm.get("name"),
            "cron": fm.get("cron"),
            "active": fm.get("active", "true"),
            "prompt": body if body.strip() else None,
            "command": fm.get("command"),
            "shell": fm.get("shell"),
            "provider": fm.get("provider"),
            "parser": fm.get("parser"),
            "parser_args": fm.get("parser_args"),
            "connector": fm.get("connector"),
            "connectors": fm.get("connectors"),
            "notify_connectors": fm.get("notify_connectors"),
        }
        _validate_task_mapping(task_data, f"{label}#task", merged_cfg)

    # 1. Directory
    if KAGE_GLOBAL_DIR.exists():
        ok(f"~/.kage/ {'ディレクトリ' if is_ja else 'Directory'}", str(KAGE_GLOBAL_DIR))
    else:
        fail(
            f"~/.kage/ {'ディレクトリ' if is_ja else 'Directory'} {t_not_found}",
            t_run_onboard,
        )

    # 2. Config
    if KAGE_CONFIG_PATH.exists():
        ok("~/.kage/config.toml", t_exists)
    else:
        warn(f"~/.kage/config.toml {t_not_found}", t_only_default)

    # 3. DB
    if KAGE_DB_PATH.exists():
        ok("kage.db", str(KAGE_DB_PATH))
    else:
        fail(f"kage.db {t_not_found}", t_run_onboard)

    # 4. Projects
    if KAGE_PROJECTS_LIST.exists():
        lines = KAGE_PROJECTS_LIST.read_text().splitlines()
        count = len([line for line in lines if line.strip()])
        ok("projects.list", f"{count} {t_proj_reg}")
    else:
        warn(f"projects.list {t_not_found}", t_run_init)

    # 5. Engine
    cfg = get_global_config()
    if cfg.default_ai_engine:
        ok("default_ai_engine", f'"{cfg.default_ai_engine}"')
    else:
        warn(t_ai_not_set, t_ai_hint)

    # 6. Provider Checks
    if cfg.default_ai_engine:
        prov = cfg.providers.get(cfg.default_ai_engine)
        if prov:
            cmd_def = cfg.commands.get(prov.command)
            if cmd_def:
                ok(
                    f"providers.{cfg.default_ai_engine}",
                    f"→ commands.{prov.command}: {cmd_def.template[0]}",
                )
            else:
                warn(
                    f"providers.{cfg.default_ai_engine}.command = '{prov.command}' {t_prov_undef}",
                    t_prov_cmd_hint,
                )
        else:
            warn(f"providers.{cfg.default_ai_engine} {t_prov_undef}", t_prov_hint)

    # 7. Validate config files and task files
    if KAGE_CONFIG_PATH.exists():
        _validate_config_file(KAGE_CONFIG_PATH, str(KAGE_CONFIG_PATH))

    try:
        from .scheduler import get_projects

        projects = get_projects()
    except Exception as e:
        projects = []
        warn("projects discovery", str(e))

    for proj_dir in sorted(projects):
        ws_cfg = proj_dir / ".kage" / "config.toml"
        if ws_cfg.exists():
            _validate_config_file(ws_cfg, str(ws_cfg))

        tasks_dir = proj_dir / ".kage" / "tasks"
        if not tasks_dir.exists():
            continue

        merged_cfg = get_global_config(workspace_dir=proj_dir)
        for task_file in sorted(list(tasks_dir.glob("*.md"))):
            _validate_task_file(task_file, merged_cfg)

    table = Table(show_header=True, header_style="bold", box=None, padding=(0, 1))
    table.add_column("", width=2)
    table.add_column(t_chk, style="bold")
    table.add_column(t_det, style="dim")
    for icon, label, detail in checks:
        table.add_row(icon, label, detail)

    console.print(table)

    fails = [c for c in checks if "✘" in c[0]]
    warns = [c for c in checks if "⚠" in c[0]]
    console.print(
        f"\n[bold]{t_res}[/bold] {len(checks)} {('項目中' if is_ja else 'items,')} [red]{len(fails)} {t_err}[/red] / [yellow]{len(warns)} {t_warn}[/yellow]"
    )
    if fails:
        raise typer.Exit(code=1)


@app.command("version")
def version():
    """Show kage version."""
    typer.echo(_resolve_version())


@app.command("skill")
def skill():
    """Fetch and display the kage agent skill definition (SKILL.md) from GitHub."""
    import urllib.request
    from rich.console import Console
    from rich.markdown import Markdown

    console = Console()
    v = _resolve_version()
    # URL pattern: https://raw.githubusercontent.com/igtm/kage/refs/tags/v0.1.2/skills/kage/SKILL.md
    url = f"https://raw.githubusercontent.com/igtm/kage/refs/tags/v{v}/skills/kage/SKILL.md"

    console.print(f"[dim]Fetching skill definition from: {url}[/dim]")
    try:
        req = urllib.request.Request(url, headers={"User-Agent": "kage-cli"})
        with urllib.request.urlopen(req) as response:
            content = response.read().decode("utf-8")
            console.print(Markdown(content))
    except Exception as e:
        console.print(f"[red]Error fetching skill definition:[/red] {e}")
        console.print(
            "[yellow]Hint:[/yellow] Make sure you are using a tagged version and have internet access."
        )

@connector_app.command("setup")
def connector_setup(
    ctype: Optional[str] = typer.Argument(None, help="Connector type (discord, slack, telegram)")
):
    """Show setup instructions for a connector type."""
    from rich.console import Console
    from rich.panel import Panel
    from rich.markdown import Markdown

    console = Console()

    if not ctype:
        console.print("[bold]Available Connector Types:[/bold]")
        console.print("- [bold magenta]discord[/bold magenta]")
        console.print("- [bold blue]slack[/bold blue]")
        console.print("- [bold cyan]telegram[/bold cyan]")
        console.print("\nRun [bold]kage connector setup discord[/bold] for instructions.")
        return

    ctype = ctype.lower()
    if ctype == "discord":
        text = """
# Discord Connector Setup Guide

1. **Create Application**: Go to [Discord Developer Portal](https://discord.com/developers/applications) and click **"New Application"**.
2. **Setup Bot**:
   - Go to the **"Bot"** tab.
   - Click **"Reset Token"** to get your **Bot Token**.
   - **CRITICAL**: Scroll down to "Privileged Gateway Intents" and enable **"Message Content Intent"**.
3. **Invite Bot**:
   - Go to **"OAuth2" -> "URL Generator"**.
   - Select scopes: `bot`.
   - Select bot permissions: `Send Messages`, `Read Messages/View Channels`, `Read Message History`.
   - Copy the generated URL and open it in your browser to invite the bot to your server.
4. **Get Channel ID**:
   - In Discord (User Settings -> Advanced), enable **"Developer Mode"**.
   - Right-click the target channel and select **"Copy Channel ID"**.
5. **Config**: Add the following to your `.kage/config.toml`:

```toml
[connectors.my_discord]
type = "discord"
poll = true   # ⚠️ Only enable in private/trusted channels (grants AI access to your PC)
bot_token = "YOUR_BOT_TOKEN"
channel_id = "YOUR_CHANNEL_ID"
system_prompt = "Optional additional instructions for this connector"
```

> **⚠️ Security**: `poll = true` allows anyone in the channel to interact with the AI, which has full access to your PC. Task notifications (via `notify_connectors`) work even with `poll = false`.
"""
        console.print(Panel(Markdown(text), title="Discord Setup", border_style="magenta"))
    elif ctype == "slack":
        text = """
# Slack Connector Setup Guide

1. **Create App**: Go to [Slack API Apps](https://api.slack.com/apps) and click **"Create New App"** (From scratch).
2. **Permissions**:
   - Go to **"OAuth & Permissions"**.
   - Scroll to **"Bot Token Scopes"** and add:
     - `channels:history`
     - `chat:write`
     - `groups:history` (if using private channels)
3. **Install**:
   - Scroll up and click **"Install to Workspace"**.
   - Copy the **"Bot User OAuth Token"** (starts with `xoxb-`).
4. **Get Channel ID**:
   - In Slack, right-click the channel name -> **"View channel details"**.
   - The Channel ID is at the very bottom (e.g., `C1234567890`).
5. **Invite Bot**: In the target channel, type `/invite @YourBotName`.
6. **Config**: Add the following to your `.kage/config.toml`:

```toml
[connectors.my_slack]
type = "slack"
poll = true   # ⚠️ Only enable in private/trusted channels (grants AI access to your PC)
bot_token = "xoxb-YOUR_TOKEN"
channel_id = "YOUR_CHANNEL_ID"
system_prompt = "Optional additional instructions for this connector"
```

> **⚠️ Security**: `poll = true` allows anyone in the channel to interact with the AI, which has full access to your PC. Task notifications (via `notify_connectors`) work even with `poll = false`.
"""
        console.print(Panel(Markdown(text), title="Slack Setup", border_style="blue"))
    elif ctype == "telegram":
        text = """
# Telegram Connector Setup Guide

1. **Create Bot**: Open Telegram and message [@BotFather](https://t.me/BotFather). Send `/newbot` and follow the prompts to create your bot.
2. **Get Bot Token**: BotFather will give you a **Bot Token** (e.g., `123456:ABC-DEF1234ghIkl-zyx57W2v1u123ew11`).
3. **Get Chat ID**:
   - Add the bot to your group or start a DM with the bot.
   - Send a message to the bot/group.
   - Open `https://api.telegram.org/bot<YOUR_BOT_TOKEN>/getUpdates` in your browser.
   - Find `"chat":{"id":...}` — this is your **Chat ID** (may be negative for groups).
4. **Config**: Add the following to your `.kage/config.toml`:

```toml
[connectors.my_telegram]
type = "telegram"
poll = true   # ⚠️ Only enable in private/trusted chats (grants AI access to your PC)
bot_token = "YOUR_BOT_TOKEN"
chat_id = "YOUR_CHAT_ID"
system_prompt = "Optional additional instructions for this connector"
```

> **⚠️ Security**: `poll = true` allows anyone in the chat to interact with the AI, which has full access to your PC. Task notifications (via `notify_connectors`) work even with `poll = false`.
"""
        console.print(Panel(Markdown(text), title="Telegram Setup", border_style="cyan"))
    else:
        console.print(f"[red]Unknown connector type: {ctype}[/red]")
        console.print("Available types: discord, slack, telegram")

if __name__ == "__main__":
    app()
@connector_app.command("list")
def connector_list():
    """List all configured connectors."""
    from .config import get_global_config
    from rich.console import Console
    from rich.table import Table

    console = Console()
    config = get_global_config()
    
    if not config.connectors:
        console.print("[yellow]No connectors configured.[/yellow]")
        console.print("Add [[connectors.name]] blocks to your config.toml.")
        return

    table = Table(show_header=True, header_style="bold magenta")
    table.add_column("Name", style="bold")
    table.add_column("Type")
    table.add_column("Status")
    table.add_column("Details")

    for name, c_dict in config.connectors.items():
        c_type = c_dict.get("type", "unknown")
        # Handle tomlkit boolean types or standard ones
        is_polling = c_dict.get("poll", False)
        if hasattr(is_polling, "unwrap"):
            is_polling = is_polling.unwrap()
            
        status = "[green]Poll ON[/green]" if is_polling else "[dim]Poll OFF[/dim]"
        
        details = []
        if c_type == "discord":
            details.append(f"Channel: {c_dict.get('channel_id', 'N/A')}")
            if c_dict.get("user_id"):
                details.append(f"User Filter: {str(c_dict.get('user_id'))}")
        elif c_type == "slack":
            details.append(f"Channel: {c_dict.get('channel_id', 'N/A')}")
            if c_dict.get("user_id"):
                details.append(f"User Filter: {str(c_dict.get('user_id'))}")
        elif c_type == "telegram":
            details.append(f"Chat: {c_dict.get('chat_id', 'N/A')}")
            if c_dict.get("user_id"):
                details.append(f"User Filter: {str(c_dict.get('user_id'))}")
        
        table.add_row(name, c_type, status, ", ".join(details))

    console.print(table)


@connector_app.command("poll")
def connector_poll():
    """Poll and reply messages for connectors with poll=true."""
    from .connectors.runner import run_connectors
    from rich.console import Console
    
    console = Console()
    console.print("[bold blue]Polling connectors...[/bold blue]")
    try:
        run_connectors()
        console.print("[green]✔ Polling completed.[/green]")
    except Exception as e:
        console.print(f"[red]✘ Polling failed: {e}[/red]")



import json
import os
import re
import shutil
import signal
import sqlite3
import subprocess
import sys
from datetime import datetime
from pathlib import Path
from typing import Optional

from .config import KAGE_DB_PATH, KAGE_GLOBAL_DIR, get_global_config
from .db import log_execution, start_execution, update_execution, get_execution_pid
from .parser import TaskDef
from .ai.chat import clean_ai_reply, get_thinking_tag


def _normalize_headless_args(cmd: list[str]) -> list[str]:
    """When running without a TTY, avoid interactive approval prompts for AI CLIs."""
    if not cmd:
        return cmd

    exe = Path(cmd[0]).name

    try:
        is_tty = sys.stdin.isatty()
    except Exception:
        is_tty = False
    if is_tty:
        return cmd

    if exe == "codex":
        try:
            exec_idx = cmd.index("exec")
        except ValueError:
            return cmd

        # Split into codex-global args and exec args
        head = cmd[:exec_idx]
        tail = cmd[exec_idx + 1 :]

        # Remove exec-scoped convenience that can prompt on non-TTY
        tail = [part for part in tail if part != "--full-auto"]

        def _drop_flag_with_value(parts: list[str], flag: str) -> list[str]:
            out: list[str] = []
            i = 0
            while i < len(parts):
                if parts[i] == flag:
                    i += 2
                    continue
                out.append(parts[i])
                i += 1
            return out

        # Ensure global non-interactive policy before subcommand.
        head = _drop_flag_with_value(head, "--ask-for-approval")
        head = _drop_flag_with_value(head, "--sandbox")
        head = head + ["--ask-for-approval", "never", "--sandbox", "workspace-write"]

        return head + ["exec"] + tail

    if exe == "claude":
        # Ensure --dangerously-skip-permissions is present for claude
        # We also prefer -p (print mode) for non-interactive execution
        new_cmd = list(cmd)
        if "-p" not in new_cmd and "--print" not in new_cmd:
            # Insert -p after 'claude' if not present
            new_cmd.insert(1, "-p")

        if "--dangerously-skip-permissions" not in new_cmd:
            new_cmd.insert(1, "--dangerously-skip-permissions")
        if "--allow-dangerously-skip-permissions" not in new_cmd:
            new_cmd.insert(1, "--allow-dangerously-skip-permissions")

        return new_cmd

    if exe == "gemini":
        # Ensure --prompt (or -p) is used for headless mode.
        if "--prompt" not in cmd and "-p" not in cmd:
            new_cmd = list(cmd)
            if len(new_cmd) > 1:
                new_cmd.insert(-1, "-p")
            else:
                new_cmd.append("-p")
            return new_cmd

    return cmd


def _get_memory_dir(project_dir: Path, task_name: str) -> Path:
    """タスクのメモリディレクトリパスを返す。ディレクトリがなければ作成する。"""
    safe_name = re.sub(r"[^a-zA-Z0-9_\-]", "_", task_name)
    memory_dir = project_dir / ".kage" / "memory" / safe_name
    memory_dir.mkdir(parents=True, exist_ok=True)
    return memory_dir


def _load_recent_memories(memory_dir: Path, max_entries: int = 5) -> str:
    """直近N日分のメモリファイルを読み込み、結合した文字列として返す。"""
    date_files = sorted(memory_dir.glob("*.json"), reverse=True)
    # task.json はメモリファイルではないので除外
    date_files = [f for f in date_files if f.name != "task.json"]
    date_files = date_files[:max_entries]
    date_files.reverse()  # 古い順に並べる

    if not date_files:
        return ""

    parts = []
    for f in date_files:
        try:
            content = f.read_text(encoding="utf-8")
            date_label = f.stem  # e.g. "2026-02-28"
            parts.append(f"--- {date_label} ---\n{content}")
        except Exception:
            continue
    return "\n\n".join(parts)


def _load_task_json(memory_dir: Path) -> dict:
    """task.json を読み込む。存在しなければ空dictを返す。"""
    task_json_path = memory_dir / "task.json"
    if task_json_path.exists():
        try:
            return json.loads(task_json_path.read_text(encoding="utf-8"))
        except Exception:
            return {}
    return {}


def _compute_prompt_hash(prompt: str) -> str:
    """プロンプト本文のSHA256ハッシュを返す。"""
    import hashlib
    return hashlib.sha256(prompt.strip().encode("utf-8")).hexdigest()


def _deactivate_task(task_file: Path):
    if not task_file.exists():
        return
    content = task_file.read_text(encoding="utf-8")
    new_content = re.sub(r"active:\s*(true|false)", "active: false", content)
    if content == new_content:
        # active が無い場合は cron: の後に追加
        new_content = re.sub(r"(cron:.*?\n)", r"\1active: false\n", content)
    task_file.write_text(new_content, encoding="utf-8")
    print(f"Task deactivated: {task_file.name}")


def _get_lock_path(project_dir: Path, task_name: str) -> Path:
    safe_proj = re.sub(r"[^a-zA-Z0-9_\-]", "_", project_dir.name)
    safe_task = re.sub(r"[^a-zA-Z0-9_\-]", "_", task_name)
    path = KAGE_GLOBAL_DIR / "locks" / f"{safe_proj}_{safe_task}.lock"
    path.parent.mkdir(parents=True, exist_ok=True)
    return path


def _check_running(lock_path: Path) -> Optional[int]:
    if not lock_path.exists():
        return None
    try:
        pid = int(lock_path.read_text().strip())
        os.kill(pid, 0)  # Check if process exists
        return pid
    except (ProcessLookupError, ValueError, OverflowError):
        return None

def _notify_connectors(task: TaskDef, status: str, stdout: str, stderr: str):
    if not task.notify_connectors:
        return
        
    from .connectors.runner import get_connector
    
    msg = f"**[{task.name}]** Execution completed with status: `{status}`"
    
    if stdout:
        msg += f"\n{stdout[:2000]}"
        if len(stdout) > 2000:
            msg += "\n...(truncated)"
        
    for c_name in task.notify_connectors:
        connector = get_connector(c_name)
        if connector:
            connector.send_message(msg)


def stop_execution(exec_id: str):
    """実行中のタスクを停止する。"""
    pid = get_execution_pid(exec_id)
    if not pid:
        print(f"No PID found for execution {exec_id}")
        return

    try:
        # プロセスグループ全体にSIGTERMを送信
        os.killpg(os.getpgid(pid), signal.SIGTERM)
        update_execution(exec_id, "STOPPED", "", "Terminated by user")
        print(f"Stopped execution {exec_id} (PID: {pid})")
    except ProcessLookupError:
        # 既に終了している場合
        update_execution(exec_id, "STOPPED", "", "Terminated by user (already dead)")
    except Exception as e:
        print(f"Failed to stop execution {exec_id}: {e}")


def execute_task(project_dir: Path, task: TaskDef, task_file: Optional[Path] = None):
    if not task.active:
        print(f"Skipping inactive task '{task.name}' in {project_dir}")
        return

    from .config import get_system_prompt
    from .parser import ConcurrencyPolicy, ExecutionMode

    # 多重起動チェック
    lock_path = _get_lock_path(project_dir, task.name)
    running_pid = _check_running(lock_path)

    if running_pid:
        if task.concurrency_policy == ConcurrencyPolicy.FORBID:
            print(
                f"Task '{task.name}' is already running (PID: {running_pid}). Skipping."
            )
            return
        elif task.concurrency_policy == ConcurrencyPolicy.REPLACE:
            print(
                f"Task '{task.name}' is already running (PID: {running_pid}). Terminating and replacing."
            )
            try:
                os.kill(running_pid, signal.SIGTERM)
            except Exception:
                pass

    global_config = get_global_config(workspace_dir=project_dir)
    system_prompt = get_system_prompt(workspace_dir=project_dir)

    # ロックファイル作成
    try:
        lock_path.write_text(str(os.getpid()))

        # メモリの読み込み（直近N件）
        memory_dir = _get_memory_dir(project_dir, task.name)
        memory_context_str = _load_recent_memories(
            memory_dir, max_entries=global_config.memory_max_entries
        )

        provider = None
        parser_type = "raw"
        p_args = ""
        cmd = []

        if task.prompt:
            # タスク管理 (task.json) の読み込みと prompt_hash チェック
            task_plan = _load_task_json(memory_dir)
            current_hash = _compute_prompt_hash(task.prompt)

            task_plan_context = ""
            if task_plan:
                stored_hash = task_plan.get("prompt_hash", "")
                plan_json = json.dumps(task_plan, indent=2, ensure_ascii=False)
                if stored_hash == current_hash:
                    task_plan_context = (
                        f"\n\n## Task Plan (.kage/memory/{re.sub(r'[^a-zA-Z0-9_-]', '_', task.name)}/task.json)\n"
                        f"{plan_json}"
                    )
                else:
                    task_plan_context = (
                        f"\n\n## Task Plan (.kage/memory/{re.sub(r'[^a-zA-Z0-9_-]', '_', task.name)}/task.json)\n"
                        f"\u26a0 **The task prompt has been updated** (hash mismatch). "
                        f"Review the previous task plan below and regenerate it based on the new instructions. "
                        f"Reuse completed work where applicable. Update `prompt_hash` to `{current_hash}`.\n\n"
                        f"Previous plan:\n{plan_json}"
                    )
            else:
                task_plan_context = (
                    f"\n\n## Task Plan\n"
                    f"No task plan found. Create one by writing to "
                    f"`.kage/memory/{re.sub(r'[^a-zA-Z0-9_-]', '_', task.name)}/task.json`.\n"
                    f"Use `prompt_hash`: `{current_hash}`"
                )

            # メモリコンテキスト
            if memory_context_str:
                memory_section = f"\n\n## Recent Memory (most recent {global_config.memory_max_entries} entries)\n{memory_context_str}"
            else:
                memory_section = (
                    f"\n\n## Recent Memory\n"
                    f"No previous memory found. You can write memory files to "
                    f"`.kage/memory/{re.sub(r'[^a-zA-Z0-9_-]', '_', task.name)}/YYYY-MM-DD.json`."
                )

            # プロバイダーの解決
            engine_name = task.provider or global_config.default_ai_engine
            if not engine_name:
                msg = "AIエンジンが未指定です。"
                print(f"[kage] ERROR: {msg}")
                log_execution(str(project_dir), task.name, "FAILED", "", msg)
                return

            # プロンプトの構築: System Prompt + Task Plan + Memory + Task Instructions
            thinking_tag = get_thinking_tag(engine_name)
            formatted_system_prompt = system_prompt.replace("{thinking_tag}", thinking_tag)
            full_prompt = f"{formatted_system_prompt}{task_plan_context}{memory_section}\n\n## Task Instructions\n{task.prompt}"

            provider = global_config.providers.get(engine_name)
            extra_args = task.ai.args if task.ai and task.ai.args else []

            resolved_template = None
            if task.command_template:
                resolved_template = task.command_template
            elif provider:
                cmd_def = global_config.commands.get(provider.command)
                if cmd_def:
                    resolved_template = cmd_def.template

            parser_type = task.parser or (provider.parser if provider else "raw")
            p_args = task.parser_args or (provider.parser_args if provider else "")

            if resolved_template:
                cmd = [
                    part.replace("{prompt}", full_prompt) for part in resolved_template
                ]
                cmd.extend(extra_args)
            else:
                cmd = [engine_name, full_prompt] + extra_args

        elif task.command:
            shell_cmd = task.shell or "sh"
            cmd = [shell_cmd, "-c", task.command]
        else:
            log_execution(
                str(project_dir),
                task.name,
                "FAILED",
                "",
                "No prompt or command specified",
            )
            return

        try:
            from .db import init_db
            init_db()  # Migration ensure
            # まだPIDがわからないのでNoneで開始
            exec_id = start_execution(str(project_dir), task.name)

            print(f"Executing task '{task.name}' in {project_dir}")
            env = os.environ.copy()
            if global_config.env_path:
                env["PATH"] = global_config.env_path

            if cmd and cmd[0]:
                exe_path = shutil.which(cmd[0], path=env.get("PATH"))
                if exe_path:
                    cmd[0] = exe_path
            cmd = _normalize_headless_args(cmd)

            # タイムアウト設定 (デフォルトなし)
            timeout = task.timeout_minutes * 60 if task.timeout_minutes else None

            # プロセス開始
            proc = subprocess.Popen(
                cmd,
                cwd=project_dir,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                env=env,
                start_new_session=True,  # 新しいセッション（プロセスグループ）を作成
            )

            # PID を更新
            conn = sqlite3.connect(KAGE_DB_PATH)
            cursor = conn.cursor()
            cursor.execute("UPDATE executions SET pid = ? WHERE id = ?", (proc.pid, exec_id))
            conn.commit()
            conn.close()

            try:
                stdout, stderr = proc.communicate(timeout=timeout)
                returncode = proc.returncode
            except subprocess.TimeoutExpired:
                # タイムアウト時はプロセスグループごと終了させる
                os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
                stdout, stderr = proc.communicate()
                returncode = -1
                raise  # re-raise to be caught by outer try

            result = subprocess.CompletedProcess(cmd, returncode, stdout, stderr)

            # autostop チェック: task.json の sub_tasks が全て done の場合
            if task.prompt and task.mode == ExecutionMode.AUTOSTOP and task_file:
                should_stop = False
                # task.json による判定
                updated_plan = _load_task_json(memory_dir)
                if updated_plan.get("sub_tasks"):
                    all_done = all(
                        t.get("status") == "done"
                        for t in updated_plan["sub_tasks"]
                    )
                    if all_done:
                        should_stop = True
                if should_stop:
                    _deactivate_task(task_file)

            if task.mode == ExecutionMode.ONCE and task_file:
                _deactivate_task(task_file)

            if task.prompt and parser_type == "jq" and p_args:
                try:
                    jq_result = subprocess.run(
                        ["jq", "-r", p_args],
                        input=result.stdout,
                        capture_output=True,
                        text=True,
                        env=env,
                    )
                    if jq_result.returncode == 0:
                        result.stdout = jq_result.stdout
                    else:
                        result.stderr += f"\n[jq err]: {jq_result.stderr}"
                except Exception as jq_e:
                    result.stderr += f"\n[jq exc]: {str(jq_e)}"

            status = "SUCCESS" if result.returncode == 0 else "FAILED"
            # Clean thinking tags from AI output before storing and notifying
            if task.prompt:
                result.stdout = clean_ai_reply(result.stdout)
            update_execution(exec_id, status, result.stdout, result.stderr)
            _notify_connectors(task, status, result.stdout, result.stderr)
        except subprocess.TimeoutExpired:
            stderr = f"Task timed out after {task.timeout_minutes} minutes"
            update_execution(
                exec_id,
                "TIMEOUT",
                "",
                stderr,
            )
            _notify_connectors(task, "TIMEOUT", "", stderr)
        except Exception as e:
            update_execution(exec_id, "ERROR", "", str(e))
            _notify_connectors(task, "ERROR", "", str(e))
    finally:
        if lock_path.exists():
            lock_path.unlink()
        
        # 最終的に PID を NULL にしておく (終了済みを明示)
        try:
            conn = sqlite3.connect(KAGE_DB_PATH)
            cursor = conn.cursor()
            cursor.execute("UPDATE executions SET pid = NULL WHERE id = ?", (exec_id,))
            conn.commit()
            conn.close()
        except:
            pass



from abc import ABC, abstractmethod

class BaseConnector(ABC):
    def __init__(self, name: str, config):
        self.name = name
        self.config = config

    @abstractmethod
    def poll_and_reply(self):
        """
        Poll for new messages from the external chat service and reply to them.
        This method will be called periodically by the kage cron (scheduler).
        """
        pass

    @abstractmethod
    def send_message(self, text: str):
        """
        Send a notification message to the external chat service.
        """
        pass



import pytest
import time
import subprocess
from pathlib import Path
from kage.executor import execute_task, stop_execution
from kage.parser import TaskDef
from kage.db import init_db, KAGE_DB_PATH
import sqlite3

@pytest.fixture
def clean_db(mocker, tmp_path):
    db_path = tmp_path / "kage.db"
    mocker.patch("kage.db.KAGE_DB_PATH", db_path)
    mocker.patch("kage.config.KAGE_DB_PATH", db_path)
    init_db()
    yield db_path

def test_stop_execution_flow(tmp_path: Path, clean_db):
    # 長時間実行されるコマンドを定義 (10秒間sleep)
    # 既存のタスクと衝突しないようにユニークな名前を使用
    task_name = f"test_task_{int(time.time())}"
    task = TaskDef(name=task_name, cron="* * * * *", command="sleep 10")
    
    # 実行を別スレッドで開始
    import threading
    thread = threading.Thread(target=execute_task, args=(tmp_path, task))
    thread.start()
    
    # PIDがDBに書き込まれるまで少し待つ
    exec_id = None
    db_path = clean_db
    for i in range(20):
        time.sleep(0.5)
        conn = sqlite3.connect(db_path)
        cursor = conn.cursor()
        cursor.execute("SELECT id, task_name, status, pid FROM executions")
        all_rows = cursor.fetchall()
        print(f"Polling {i}: {all_rows}")
        
        for r_id, r_name, r_status, r_pid in all_rows:
            if r_name == task_name and r_status == 'RUNNING' and r_pid:
                exec_id = r_id
                break
        conn.close()
        if exec_id:
            break
    
    assert exec_id is not None, "Execution should be started and PID stored"
    
    # 停止実行
    stop_execution(exec_id)
    
    # 終了を待つ
    thread.join(timeout=5)
    
    # ステータス確認
    conn = sqlite3.connect(db_path)
    cursor = conn.cursor()
    cursor.execute("SELECT status, stderr FROM executions WHERE id = ?", (exec_id,))
    row = cursor.fetchone()
    conn.close()
    
    assert row[0] == "STOPPED"
    assert "Terminated by user" in row[1]
